{% if router == 'fastapi' %}
    {% include '_fastapi_app.py.jinja' %}
{% elif router == 'starlette' %}
    {% include '_starlette_app.py.jinja' %}
{% elif router == 'blacksheep' %}
    {% include '_blacksheep_app.py.jinja' %}
{% elif router == 'litestar' %}
    {% include '_litestar_app.py.jinja' %}
{% endif %}
